# Copyright Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ETOS API handler."""
import time
import logging
from typing import Union
from json import JSONDecodeError

from urllib3.exceptions import MaxRetryError, NewConnectionError
import requests
from requests.exceptions import HTTPError

from .schema import RequestSchema, ResponseSchema


class ETOS:  # pylint:disable=too-few-public-methods
    """Handle communication with ETOS."""

    logger = logging.getLogger(__name__)
    reason = ""

    def __init__(self, cluster: str) -> None:
        """Initialize ETOS."""
        self.cluster = cluster

    def start(self, request_data: RequestSchema) -> Union[ResponseSchema, None]:
        """Start ETOS."""
        self.logger.info("Check connection to ETOS.")
        if not self.__check_connection():
            self.reason = "Unable to connect to ETOS. Please check your connection."
            return None
        self.logger.info("Connection successful.")
        self.logger.info("Triggering ETOS.")
        response = self.__start(request_data)
        if not response:
            self.reason = "Failed to start ETOS"
            return None
        return response

    def __check_connection(self) -> bool:
        """Check connection to ETOS."""
        try:
            response = requests.get(f"{self.cluster}/selftest/ping", timeout=5)
            response.raise_for_status()
            return True
        except Exception:  # pylint:disable=broad-exception-caught
            return False

    def __start(self, request_data: RequestSchema) -> Union[ResponseSchema, None]:
        """Start an ETOS testrun."""
        response = self.__retry_trigger_etos(request_data)
        if not response:
            return None

        return response

    def __retry_trigger_etos(
        self, request_data: RequestSchema
    ) -> Union[ResponseSchema, None]:
        """Trigger ETOS, retrying on non-client errors until successful."""
        end_time = time.time() + 30
        while time.time() < end_time:
            response = requests.post(
                f"{self.cluster}/etos", json=request_data.dict(), timeout=10
            )
            if self.__should_retry(response):
                time.sleep(2)
                continue
            if not response.ok:
                return None
            return ResponseSchema.from_response(response.json())
        self.logger.critical("Failed to trigger ETOS.")
        return None

    def __should_retry(self, response: requests.Response) -> bool:
        """Check response to see whether it is worth retrying or not."""
        try:
            response.raise_for_status()
        except HTTPError as http_error:
            if 400 <= http_error.response.status_code < 500:
                try:
                    response_json = response.json()
                except JSONDecodeError:
                    self.logger.info("Raw response from ETOS: %r", response.text)
                    response_json = {"detail": "Unknown client error from ETOS"}
                self.logger.critical(response_json.get("detail"))
                return False
            return True
        except (
            ConnectionError,
            NewConnectionError,
            MaxRetryError,
            TimeoutError,
        ):
            self.logger.exception("Network connectivity errors when triggering ETOS.")
            return True
        return False



# Copyright Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""GraphQL query handler."""
from typing import Iterator, Optional
from etos_lib import ETOS
from .graphql_queries import (
    ARTIFACTS,
    ACTIVITY_TRIGGERED,
    ACTIVITY_FINISHED,
    ACTIVITY_CANCELED,
    MAIN_TEST_SUITES_STARTED,
    TEST_SUITE_STARTED,
    TEST_SUITE_FINISHED,
    TEST_SUITE,
    ANNOUNCEMENTS,
    ENVIRONMENTS,
)


class Search(dict):
    """A search dictionary for eiffel graphql API.

    This makes it easier to create search strings that are more
    coherent. Just create them as you would do a dictionary.
    When this object is then converted to a string in the graphql
    query it will convert the dictionary into a query string that
    works with the eiffel graphql API.
    """

    def __str__(self) -> str:
        """Convert dictionary to an eiffel graphql API compatible string."""
        strings = []
        for key, value in self.items():
            if key == "search":
                strings.append(f'search: "{self["search"]}"')
            else:
                strings.append(f"{key}: {value}")
        return ", ".join(strings)


def request(etos: ETOS, query: str, search: Search) -> Iterator[dict]:
    """Request graphql in a generator."""
    wait_generator = etos.utils.wait(
        etos.graphql.execute, query=query % search, timeout=60
    )
    yield from wait_generator


def search_for(etos: ETOS, query: str, search: Search, node: str) -> Iterator[dict]:
    """Request graphql and search for node in response."""
    for response in request(etos, query, search):
        if response:
            for _, event in etos.graphql.search_for_nodes(response, node):
                yield event
            return None  # StopIteration
    return None  # StopIteration


def request_all(etos: ETOS, query: str, search: Search, node: str) -> Iterator[dict]:
    """Request all events for a given query."""
    limit = 100
    search["last"] = limit
    while True:
        last_event = None
        count = 0
        for event in search_for(etos, query, search, node):
            if event.get("meta", {}).get("time") is None:
                raise AssertionError(
                    "Meta time must be added to graphql query in order"
                    " to use this function"
                )
            last_event = event
            count += 1
            yield event
        if last_event is None:
            return None  # StopIteration
        if count < limit:
            return None  # StopIteration
        search["search"]["meta.time"] = {"$lt": last_event["meta"]["time"]}


def get_one(etos: ETOS, query: str, search: Search, node: str) -> Optional[dict]:
    """Request graphql and get a single node from response."""
    for response in request(etos, query, search):
        if response:
            try:
                _, event = next(etos.graphql.search_for_nodes(response, node))
            except StopIteration:
                return None
            return event
    return None


def request_suite(etos: ETOS, suite_id: str) -> Optional[dict]:
    """Request a tercc from graphql."""
    return get_one(
        etos,
        TEST_SUITE,
        Search(search={"meta.id": suite_id}),
        "testExecutionRecipeCollectionCreated",
    )


def request_activity(etos: ETOS, suite_id: str) -> Optional[dict]:
    """Request an activity event from graphql."""
    return get_one(
        etos,
        ACTIVITY_TRIGGERED,
        Search(search={"links.type": "CAUSE", "links.target": suite_id}),
        "activityTriggered",
    )


def request_activity_finished(etos: ETOS, activity_id: str) -> Optional[dict]:
    """Request an activity finished event from graphql."""
    return get_one(
        etos,
        ACTIVITY_FINISHED,
        Search(
            search={"links.type": "ACTIVITY_EXECUTION", "links.target": activity_id}
        ),
        "activityFinished",
    )


def request_activity_canceled(etos: ETOS, activity_id: str) -> Optional[dict]:
    """Request an activity event from graphql."""
    return get_one(
        etos,
        ACTIVITY_CANCELED,
        Search(
            search={"links.type": "ACTIVITY_EXECUTION", "links.target": activity_id}
        ),
        "activityCanceled",
    )


def request_sub_test_suite_started(etos: ETOS, main_suite_id: str) -> Iterator[dict]:
    """Request test suite started from graphql."""
    yield from search_for(
        etos,
        TEST_SUITE_STARTED,
        Search(search={"links.type": "CAUSE", "links.target": main_suite_id}),
        "testSuiteStarted",
    )


def request_main_test_suites_started(etos: ETOS, activity_id: str) -> Iterator[dict]:
    """Request test suite started from graphql."""
    yield from search_for(
        etos,
        MAIN_TEST_SUITES_STARTED,
        Search(
            search={
                "links.type": "CONTEXT",
                "links.target": activity_id,
                "data.categories": {"$ne": "Sub suite"},
            }
        ),
        "testSuiteStarted",
    )


def request_test_suite_finished(etos: ETOS, test_suite_id: str) -> Optional[dict]:
    """Request test suite finished from graphql."""
    return get_one(
        etos,
        TEST_SUITE_FINISHED,
        Search(
            search={
                "links.type": "TEST_SUITE_EXECUTION",
                "links.target": test_suite_id,
            },
            last=1,
        ),
        "testSuiteFinished",
    )


def request_announcements(etos: ETOS, ids: list[str]) -> Iterator[dict]:
    """Request announcements from graphql."""
    yield from search_for(
        etos,
        ANNOUNCEMENTS,
        Search(search={"$or": [{"links.target": _id} for _id in ids]}),
        "announcementPublished",
    )


def request_environment(etos: ETOS, ids: list[str]) -> Iterator[dict]:
    """Request environments from graphql."""
    yield from request_all(
        etos,
        ENVIRONMENTS,
        Search(search={"$or": [{"links.target": _id} for _id in ids]}),
        "environmentDefined",
    )


def request_artifacts(etos: ETOS, context: str) -> Iterator[dict]:
    """Request artifacts from graphql."""
    yield from request_all(
        etos,
        ARTIFACTS,
        Search(search={"links.type": "CONTEXT", "links.target": context}),
        "artifactCreated",
    )



# Copyright Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Announcer module."""
import logging
from etos_client.events.events import Events, TestSuite, Announcement


class Announcer:  # pylint:disable=too-few-public-methods
    """Announce the state of ETOS."""

    logger = logging.getLogger(__name__)

    def __init__(self):
        """Init."""
        self.__already_logged = []

    def __log_announcements(self, announcements: list[Announcement]) -> None:
        """Get latest new announcements."""
        for announcement in announcements:
            if announcement not in self.__already_logged:
                self.__already_logged.append(announcement)
                self.logger.info("%s: %s", announcement.heading, announcement.body)

    def __etos_state_information(self, test_suites: list[TestSuite]) -> str:
        """Generate text based on ETOS state."""
        message_template = (
            "{announcement}\t"
            "Started : {started_length}\t"
            "Finished: {finished_length}\t"
        )
        try:
            announcement = self.__already_logged[-1].body
        except (KeyError, IndexError, TypeError):
            announcement = ""

        started_length = 0
        finished_length = 0
        for test_suite in test_suites:
            for sub_suite in test_suite.sub_suites:
                started_length += 1
                if sub_suite.finished:
                    finished_length += 1

        params = {
            "started_length": started_length,
            "finished_length": finished_length,
            "announcement": announcement,
        }
        return message_template.format(**params)

    def announce(self, events: Events) -> None:
        """Announce the ETOS state."""
        if not events.tercc:
            return
        if not events.activity.triggered:
            return
        self.__log_announcements(events.announcements)
        if events.main_suites:
            self.logger.info(self.__etos_state_information(events.main_suites))



# Copyright Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ETOS request data."""
import json
from uuid import UUID
from typing import Optional, Union

from pydantic import BaseModel, validator

# Pydantic requires validators first argument to be cls and the methods cannot be classmethods
# pylint:disable=no-self-argument


class RequestSchema(BaseModel):
    """Request model for the ETOS start API."""

    artifact_id: Optional[str]
    artifact_identity: Optional[str]
    test_suite_url: str
    dataset: Optional[Union[dict, list]] = {}
    execution_space_provider: Optional[str] = "default"
    iut_provider: Optional[str] = "default"
    log_area_provider: Optional[str] = "default"

    @classmethod
    def from_args(cls, args: list[str]) -> "RequestSchema":
        """Create a RequestSchema from an argument list from argparse."""
        return RequestSchema(
            artifact_id=args.identity,
            artifact_identity=args.identity,
            test_suite_url=args.test_suite,
            dataset=args.dataset,
            execution_space_provider=args.execution_space_provider,
            iut_provider=args.iut_provider,
            log_area_provider=args.log_area_provider,
        )

    @validator("artifact_identity", always=True)
    def trim_identity_if_necessary(
        cls, artifact_identity: Optional[str], values
    ) -> Optional[str]:
        """Trim identity if id is set."""
        if values.get("artifact_id") is not None:
            return None
        return artifact_identity

    @validator("artifact_id", pre=True)
    def is_uuid(cls, artifact_id: Optional[str]) -> Optional[str]:
        """Test if string is a valid UUID v4."""
        try:
            UUID(artifact_id, version=4)
        except ValueError:
            return None
        return artifact_id

    @validator("dataset", always=True)
    def dataset_list_trimming(cls, dataset: Optional[Union[dict, list]]) -> list[dict]:
        """Trim away list completely should the dataset only contain a single index."""
        if dataset is None:
            return {}
        if len(dataset) == 1:
            return json.loads(dataset[0])
        return [json.loads(data) for data in dataset]


class ResponseSchema(BaseModel):
    """Response model for the ETOS start API."""

    event_repository: str
    tercc: UUID
    artifact_id: UUID
    artifact_identity: str

    @classmethod
    def from_response(cls, response: dict) -> "ResponseSchema":
        """Create a ResponseSchema from a dictionary. Typically used for ETOS API http response."""
        return ResponseSchema(
            event_repository=response.get("event_repository"),
            tercc=response.get("tercc"),
            artifact_id=response.get("artifact_id"),
            artifact_identity=response.get("artifact_identity"),
        )



# Copyright Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ETOS client event collector."""
from uuid import UUID
from typing import Optional
from .events import (
    Events,
    Activity,
    TestSuite,
    SubSuite,
    Environment,
    Announcement,
    Artifact,
)


class Collector:  # pylint:disable=too-few-public-methods
    """Collect events from an event repository."""

    def __init__(
        self,
        etos_library: "etos_lib.ETOS",
        event_repository: "etos_client.event_repository.graphql",
    ) -> None:
        """Initialize with event repository."""
        self.event_repository = event_repository
        self.etos_library = etos_library
        self.__events = Events()

    def __tercc(self, tercc_id: UUID) -> dict:
        """Get TERCC event to make sure it has been sent."""
        if self.__events.tercc is None:
            return self.event_repository.request_suite(self.etos_library, str(tercc_id))
        return self.__events.tercc

    def __main_test_suites(self, activity_id: UUID) -> list[TestSuite]:
        """Collect main test suite events for an ETOS activity."""
        test_suites = []
        collected = {
            suite.started["meta"]["id"]: suite for suite in self.__events.main_suites
        }
        for started in self.event_repository.request_main_test_suites_started(
            self.etos_library, str(activity_id)
        ):
            if started["meta"]["id"] not in collected.keys():
                collected[started["meta"]["id"]] = TestSuite(started=started)
        for test_suite_id, test_suite in collected.items():
            if test_suite.finished is None:
                test_suite.finished = self.event_repository.request_test_suite_finished(
                    self.etos_library, test_suite_id
                )
            test_suites.append(test_suite)
        return test_suites

    def __sub_test_suites(self, main_suite: TestSuite) -> list[SubSuite]:
        """Collect sub suites for a single main suite."""
        sub_suites = []
        collected = {
            suite.started["meta"]["id"]: suite for suite in main_suite.sub_suites
        }
        for started in self.event_repository.request_sub_test_suite_started(
            self.etos_library, main_suite.started["meta"]["id"]
        ):
            if started["meta"]["id"] not in collected:
                collected[started["meta"]["id"]] = SubSuite(started=started)
        for test_suite_id, test_suite in collected.items():
            if test_suite.finished is None:
                test_suite.finished = self.event_repository.request_test_suite_finished(
                    self.etos_library, test_suite_id
                )
            sub_suites.append(test_suite)
        return sub_suites

    def __environments(
        self, activity_id: UUID, test_suites: list[TestSuite]
    ) -> list[Environment]:
        """Collect environment defined events from an ETOS test run."""
        ids = [str(activity_id)]
        for test_suite in test_suites:
            ids.append(test_suite.started["meta"]["id"])
            for sub_suite in test_suite.sub_suites:
                ids.append(sub_suite.started["meta"]["id"])
        environments = []
        for environment in self.event_repository.request_environment(
            self.etos_library, ids
        ):
            environments.append(
                Environment(
                    name=environment["data"]["name"], uri=environment["data"]["uri"]
                )
            )
        return environments

    def __activity(self, tercc_id: UUID) -> Activity:
        """Collect activity events from an ETOS test run."""
        activity = Activity()
        triggered = self.__events.activity.triggered
        if triggered is None:
            triggered = self.event_repository.request_activity(
                self.etos_library, str(tercc_id)
            )
        activity.triggered = triggered
        if activity.triggered is None:
            return activity

        canceled = self.__events.activity.canceled
        if canceled is None:
            canceled = self.event_repository.request_activity_canceled(
                self.etos_library, triggered["meta"]["id"]
            )
        activity.canceled = canceled

        finished = self.__events.activity.finished
        if finished is None:
            finished = self.event_repository.request_activity_finished(
                self.etos_library, triggered["meta"]["id"]
            )
        activity.finished = finished
        return activity

    def __announcements(
        self, tercc_id: UUID, activity_id: Optional[UUID]
    ) -> list[Announcement]:
        """Collect announcements for an ETOS test run."""
        ids = [str(tercc_id)]
        if activity_id is not None:
            ids.append(str(activity_id))
        announcements = []
        for announcement in self.event_repository.request_announcements(
            self.etos_library, ids
        ):
            announcements.append(Announcement.parse_obj(announcement["data"]))
        return announcements

    def __artifacts(self, activity_id: UUID) -> list[Artifact]:
        """Collect artifacts from ETOS."""
        artifacts = []
        for artifact_created in self.event_repository.request_artifacts(
            self.etos_library, str(activity_id)
        ):
            file_names = [
                _file["name"] for _file in artifact_created["data"]["fileInformation"]
            ]
            suite_name = ""
            for link in artifact_created.get("links", []):
                try:
                    suite_name = link["links"]["data"]["name"]
                except KeyError:
                    pass
            for _, location in self.etos_library.utils.search(artifact_created, "uri"):
                artifacts.append(
                    Artifact(files=file_names, suite_name=suite_name, location=location)
                )
        return artifacts

    def collect(self, tercc_id: UUID) -> Events:
        """Collect events from ETOS."""
        activity_id = None
        self.__events.tercc = self.__tercc(tercc_id)
        if self.__events.tercc is None:
            return self.__events
        self.__events.announcements = self.__announcements(tercc_id, activity_id)

        self.__events.activity = self.__activity(tercc_id)
        if self.__events.activity.triggered is None:
            return self.__events
        activity_id = UUID(self.__events.activity.triggered["meta"]["id"], version=4)
        self.__events.main_suites = self.__main_test_suites(activity_id)
        for main_suite in self.__events.main_suites:
            main_suite.sub_suites = self.__sub_test_suites(main_suite)
        self.__events.environments = self.__environments(
            activity_id, self.__events.main_suites
        )
        self.__events.artifacts = self.__artifacts(activity_id)
        return self.__events



# Copyright Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ETOS Client log downloader module."""
import logging
import time
import traceback
from pathlib import Path
from typing import Union, Optional
from threading import Thread, Lock
from multiprocessing.pool import ThreadPool
from queue import Queue, Empty
from json import JSONDecodeError

from pydantic import BaseModel
from urllib3.exceptions import MaxRetryError, NewConnectionError
import requests
from requests.exceptions import HTTPError

from etos_client.events.events import Artifact, TestSuite, SubSuite


class Downloadable(BaseModel):
    """Represent a downloadable file."""

    uri: str
    name: Optional[Path]


class LogDownloader(Thread):
    """Log downloader for ETOS client."""

    logger = logging.getLogger(__name__)

    def __init__(self) -> None:
        """Init."""
        super().__init__()
        self.__download_queue = Queue()
        self.__queued = []
        self.__exit = False
        self.__clear_queue = True
        self.__lock = Lock()
        self.failed = False

    def __retry_download(self, item: Downloadable) -> None:
        """Download files."""
        self.logger.info("Downloading %r", item)
        end_time = time.time() + 30
        while time.time() < end_time:
            response = requests.get(item.uri, stream=True, timeout=10)
            self.logger.debug(response)
            if self.__should_retry(response):
                self.logger.warning("Download failed. Retrying..")
                time.sleep(2)
                continue
            if not response.ok:
                with self.__lock:
                    self.failed = True
                self.logger.critical("Failed to download %r", item)
                return
            self.__save_file(item, response)
            return
        with self.__lock:
            self.failed = True
        self.logger.critical("Failed to download %r", item)
        return

    def __save_file(self, item: Downloadable, response: requests.Response) -> None:
        """Save downloaded file data to disk."""
        download_name = item.name
        index = 0
        while download_name.exists():
            index += 1
            download_name = download_name.with_name(f"{index}_{item.name.name}")
        self.logger.info("Saving file %s", download_name)
        with open(download_name, "wb+") as report:
            for chunk in response:
                report.write(chunk)

    def __should_retry(self, response: requests.Response) -> bool:
        """Check response to see whether it is worth retrying or not."""
        try:
            response.raise_for_status()
        except HTTPError as http_error:
            if 400 <= http_error.response.status_code < 500:
                try:
                    response_json = response.json()
                except JSONDecodeError:
                    self.logger.info("Raw response from download: %r", response.text)
                    response_json = {
                        "detail": "Unknown client error when downloading files from log area"
                    }
                self.logger.critical(response_json)
                return False
            return True
        except (
            ConnectionError,
            NewConnectionError,
            MaxRetryError,
            TimeoutError,
        ):
            self.logger.exception(
                "Network connectivity error when downloading logs and artifacts."
            )
            return True
        return False

    def __trigger_download(self, pool: ThreadPool) -> bool:
        """Get downloadable from queue and add it to threadpool."""
        try:
            item = self.__download_queue.get_nowait()
            pool.apply_async(
                self.__retry_download,
                error_callback=self.__print_traceback,
                args=(item,),
            )
            return True
        except Empty:
            return False

    def __print_traceback(self, value: Exception) -> None:
        """Print a traceback if error in pool."""
        try:
            raise value
        except:  # pylint:disable=bare-except
            with self.__lock:
                self.failed = True
            traceback.print_exc()

    def run(self) -> None:
        """Run the log downloader thread."""
        with ThreadPool() as pool:
            while True:
                if self.__exit and not self.__clear_queue:
                    self.logger.warning("Forced to exit without clearing the queue.")
                    return
                time.sleep(0.1)
                if not self.__trigger_download(pool) and self.__exit:
                    self.logger.info("Download queue empty. Exiting.")
                    return

    def stop(self, clear_queue: bool = True) -> None:
        """Stop the downloader thread.

        If clear_queue is set to False, then the log downloader will exit ASAP,
        but will wait for ongoing downloads to finish.
        """
        self.__exit = True
        self.__clear_queue = clear_queue

    def __queue_download(self, item: Downloadable) -> None:
        """Queue a downloadable for download."""
        if item.uri not in self.__queued:
            self.__download_queue.put_nowait(item)
            self.__queued.append(item.uri)

    def download_artifacts(self, artifacts: list[Artifact], path: Path) -> None:
        """Download artifacts to a path."""
        for artifact in artifacts:
            for _file in artifact.files:
                filepath = path.joinpath(f"{artifact.suite_name}_{_file}").relative_to(
                    Path.cwd()
                )
                self.__queue_download(
                    Downloadable(uri=f"{artifact.location}/{_file}", name=filepath)
                )

    def download_logs(
        self, test_suites: Union[list[TestSuite], list[SubSuite]], path: Path
    ) -> None:
        """Download logs from test suites to a path."""
        for test_suite in test_suites:
            if not test_suite.finished:
                return
            data = test_suite.finished.get("data", {})
            logs = data.get("testSuitePersistentLogs", [])
            for log in logs:
                filepath = path.joinpath(log["name"]).relative_to(Path.cwd())
                self.__queue_download(Downloadable(uri=log["uri"], name=filepath))
            if isinstance(test_suite, TestSuite):
                self.download_logs(test_suite.sub_suites, path)



# Copyright Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""GraphQL queries."""

TEST_SUITE = """
{
  testExecutionRecipeCollectionCreated(%s) {
    edges {
      node {
        data {
          batchesUri
        }
      }
    }
  }
}
"""


ACTIVITY_TRIGGERED = """
{
  activityTriggered(%s) {
    edges {
      node {
        meta {
          id
        }
      }
    }
  }
}
"""


ACTIVITY_FINISHED = """
{
  activityFinished(%s) {
    edges {
      node {
        data {
          activityOutcome {
            description
            conclusion
          }
        }
      }
    }
  }
}
"""


ACTIVITY_CANCELED = """
{
  activityCanceled(%s) {
    edges {
      node {
        data {
          reason
        }
      }
    }
  }
}
"""


TEST_SUITE_STARTED = """
{
  testSuiteStarted(%s) {
    edges {
      node {
        meta {
          id
        }
      }
    }
  }
}
"""


MAIN_TEST_SUITES_STARTED = """
{
  testSuiteStarted(%s) {
    edges {
      node {
        meta {
          id
        }
      }
    }
  }
}
"""


TEST_SUITE_FINISHED = """
{
  testSuiteFinished(%s) {
    edges {
      node {
        data {
          testSuitePersistentLogs {
            name
            uri
          }
          testSuiteOutcome {
            verdict
          }
        }
      }
    }
  }
}
"""


ANNOUNCEMENTS = """
{
  announcementPublished(%s) {
    edges {
      node {
        data {
          heading
          body
        }
      }
    }
  }
}
"""


ENVIRONMENTS = """
{
  environmentDefined(%s) {
    edges {
      node {
        data {
          name
          uri
        }
        meta {
          time
        }
      }
    }
  }
}
"""


ARTIFACTS = """
{
  artifactCreated(%s) {
    edges {
      node {
        data {
          fileInformation {
            name
          }
        }
        meta {
          time
        }
        links {
          ... on Cause {
            links {
              ... on TestSuiteStarted {
                data {
                  name
                }
              }
            }
          }
        }
        reverse {
          edges {
            node {
              ... on ArtifactPublished {
                data {
                  locations {
                    uri
                  }
                }
              }
            }
          }
        }
      }
    }
  }
}
"""



# Copyright Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ETOS client event schemas."""
from typing import Optional
from pydantic import BaseModel


class Activity(BaseModel):
    """ETOS activity events."""

    triggered: Optional[dict]
    canceled: Optional[dict]
    finished: Optional[dict]


class Environment(BaseModel):
    """ETOS environment events."""

    name: str
    uri: Optional[str]


class Announcement(BaseModel):
    """ETOS announcements."""

    heading: str
    body: str


class Artifact(BaseModel):
    """ETOS artifact."""

    files: list[str]
    suite_name: str
    location: str


class SubSuite(BaseModel):
    """ETOS sub suite events."""

    started: Optional[dict]
    finished: Optional[dict]


class TestSuite(BaseModel):
    """ETOS main suite events."""

    started: Optional[dict]
    finished: Optional[dict]
    sub_suites: list[SubSuite] = []


class Events(BaseModel):
    """ETOS events."""

    activity: Activity = Activity()
    tercc: Optional[dict]
    main_suites: list[TestSuite] = []
    environments: list[Environment] = []
    announcements: list[Announcement] = []
    artifacts: list[Artifact] = []



# -*- coding: utf-8 -*-
# Copyright Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Main executable module."""
import argparse
import sys
import time
import os
import logging
import warnings
import shutil
from pathlib import Path

from etos_lib import ETOS as ETOSLibrary
from etos_client import __version__
from etos_client.events.collector import Collector
from etos_client.etos.schema import RequestSchema
from etos_client.etos import ETOS
from etos_client.test_results import TestResults
from etos_client.announcer import Announcer
from etos_client.logs import LogDownloader
from etos_client.event_repository import graphql

LOGGER = logging.getLogger(__name__)
HOUR = 3600


def environ_or_required(key: str) -> dict:
    """Get key from environment or make it a required input.

    If a key is available as an environment variable, use it as default.
    Else require the argument to be set.

    Note:
        It is still possible to override the value with input even if the
        environment variable is set.
    """
    if os.getenv(key):
        return {"default": os.getenv(key)}
    return {"required": True}


def parse_args(args: list[str]) -> argparse.Namespace:
    """Parse command line parameters."""
    parser = argparse.ArgumentParser(
        description="Client for executing test automation suites in ETOS"
    )
    parser.add_argument(
        "cluster",
        help="Cluster is the URL to the ETOS API.",
    )
    parser.add_argument(
        "-i",
        "--identity",
        help="Artifact created identity purl or ID to execute test suite on.",
        **environ_or_required("IDENTITY"),
    )
    parser.add_argument(
        "-s",
        "--test-suite",
        help="Test suite execute. Either URL or name.",
        **environ_or_required("TEST_SUITE"),
    )

    parser.add_argument(
        "--no-tty", help="This parameter is no longer in use.", action="store_true"
    )
    parser.add_argument(
        "-w",
        "--workspace",
        default=os.getenv("WORKSPACE", os.getcwd()),
        help="Which workspace to do all the work in.",
    )
    parser.add_argument(
        "-a",
        "--artifact-dir",
        default="artifacts",
        help="Where test artifacts should be stored. Relative to workspace.",
    )
    parser.add_argument(
        "-r",
        "--report-dir",
        default="reports",
        help="Where test reports should be stored. Relative to workspace.",
    )
    parser.add_argument(
        "-d",
        "--download-reports",
        default=None,
        help="This parameter is no longer in use.",
    )

    parser.add_argument(
        "--iut-provider", default="default", help="Which IUT provider to use."
    )
    parser.add_argument(
        "--execution-space-provider",
        default="default",
        help="Which execution space provider to use.",
    )
    parser.add_argument(
        "--log-area-provider", default="default", help="Which log area provider to use."
    )
    parser.add_argument(
        "--dataset",
        action="append",
        help=(
            "Additional dataset information to the environment provider. "
            "Check with your provider which information can be supplied."
        ),
    )

    parser.add_argument(
        "--version",
        action="version",
        version=f"etos_client {__version__}",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="loglevel",
        help="set loglevel to INFO",
        action="store_const",
        default=logging.WARNING,
        const=logging.INFO,
    )
    parser.add_argument(
        "-vv",
        "--very-verbose",
        dest="loglevel",
        help="set loglevel to DEBUG",
        action="store_const",
        default=logging.WARNING,
        const=logging.DEBUG,
    )
    return parser.parse_args(args)


def setup_logging(loglevel: int) -> None:
    """Set up basic logging."""
    logformat = "[%(asctime)s] %(levelname)s:%(message)s"
    logging.basicConfig(
        level=loglevel, stream=sys.stdout, format=logformat, datefmt="%Y-%m-%d %H:%M:%S"
    )


def main(args: list[str]) -> None:  # pylint:disable=too-many-statements
    """Entry point allowing external calls."""
    args = parse_args(args)
    if args.download_reports:
        warnings.warn(
            "The '-d/--download-reports' parameter is deprecated", DeprecationWarning
        )
    if args.no_tty:
        warnings.warn("The '--no-tty' parameter is deprecated", DeprecationWarning)

    setup_logging(args.loglevel)

    artifact_dir = Path(args.workspace).joinpath(args.artifact_dir)
    artifact_dir.mkdir(exist_ok=True)
    report_dir = Path(args.workspace).joinpath(args.report_dir)
    report_dir.mkdir(exist_ok=True)

    LOGGER.info("Running in cluster: %r", args.cluster)
    etos_library = ETOSLibrary("ETOS Client", os.getenv("HOSTNAME"), "ETOS Client")
    collector = Collector(etos_library, graphql)

    etos = ETOS(args.cluster)
    response = etos.start(RequestSchema.from_args(args))
    if not response:
        sys.exit(etos.reason)

    LOGGER.info("Suite ID: %s", response.tercc)
    LOGGER.info("Artifact ID: %s", response.artifact_id)
    LOGGER.info("Purl: %s", response.artifact_identity)
    os.environ["ETOS_GRAPHQL_SERVER"] = response.event_repository
    LOGGER.info("Event repository: %r", response.event_repository)

    test_results = TestResults()
    log_downloader = LogDownloader()
    announcer = Announcer()

    clear_queue = True
    log_downloader.start()
    try:
        timeout = time.time() + HOUR * 24
        while time.time() < timeout:
            events = collector.collect(response.tercc)
            if events.activity.canceled:
                sys.exit(events.activity.canceled["data"]["reason"])
            if events.activity.finished:
                break
            announcer.announce(events)
            if events.main_suites:
                log_downloader.download_logs(events.main_suites, report_dir)
            log_downloader.download_artifacts(events.artifacts, artifact_dir)
            time.sleep(10)

        events = collector.collect(response.tercc)
        if events.main_suites:
            log_downloader.download_logs(events.main_suites, report_dir)
        log_downloader.download_artifacts(events.artifacts, artifact_dir)
    except SystemExit:
        clear_queue = False
        raise
    finally:
        log_downloader.stop(clear_queue)
        log_downloader.join()

    LOGGER.info("Archiving reports.")
    shutil.make_archive(
        artifact_dir.joinpath("reports").relative_to(Path.cwd()), "zip", report_dir
    )
    LOGGER.info("Reports: %s", report_dir)
    LOGGER.info("Artifacs: %s", artifact_dir)

    if log_downloader.failed:
        sys.exit("ETOS logs did not download successfully.")

    result, message = test_results.get_results(events)
    if result:
        LOGGER.info(message)
    else:
        LOGGER.error(message)


def run() -> None:
    """Entry point for console_scripts."""
    main(sys.argv[1:])


if __name__ == "__main__":
    run()



# Copyright Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ETOS test result handler."""
import logging
from typing import Optional

from etos_client.events.events import Events, TestSuite, SubSuite

# pylint:disable=too-few-public-methods


class TestResults:
    """Get test results from an ETOS testrun."""

    logger = logging.getLogger(__name__)

    def __has_failed(self, test_suites: list[TestSuite]) -> bool:
        """Check if any test suite has failed in a list of test suites."""
        for test_suite in test_suites:
            outcome = test_suite.finished["data"]["testSuiteOutcome"]
            if outcome["verdict"] != "PASSED":
                return True
        return False

    def __count_sub_suite_failures(self, test_suites: list[SubSuite]) -> int:
        """Count the number of sub suite failures in a list of sub suites."""
        failures = 0
        for sub_suite in test_suites:
            outcome = sub_suite.finished["data"]["testSuiteOutcome"]
            if outcome["verdict"] != "PASSED":
                failures += 1
        return failures

    def __test_result(self, test_suites: list[TestSuite]) -> tuple[bool, str]:
        """Build test results based on events retrieved."""
        if not self.__has_failed(test_suites):
            return True, "Test suite finished successfully."

        failures = 0
        sub_suites = 0
        for test_suite in test_suites:
            failures += self.__count_sub_suite_failures(test_suite.sub_suites)
            sub_suites += len(test_suite.sub_suites)
        return False, f"{failures}/{sub_suites} test suites failed."

    def get_results(self, events: Events) -> tuple[Optional[bool], Optional[str]]:
        """Get results from an ETOS testrun."""
        if not events.tercc:
            return None, None
        if not events.activity.triggered:
            return None, None
        if events.activity.canceled:
            return None, None
        if not events.main_suites:
            return None, None
        for test_suite in events.main_suites:
            if not test_suite.finished:
                return None, None
        return self.__test_result(events.main_suites)
